/*
 * songs - a little tool to record and mix audio files.
 *
 * public domain
 *
 * http://sed.free.fr/songs
 */

#include "eq.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

/* don't ask me why with -std=c99 there is no M_PI on this debian... */
#ifndef M_PI
#define M_PI	3.14159265358979323846  /* pi */
#endif

#include "error.h"
#include "eq_gui.h"
#include "fft.h"

static double do_fir_left(eq *e, double sample)
{
  int i;
  int cur = e->pos_in_delay_left;
  double ret = 0;

  e->delay_left[e->pos_in_delay_left] = sample;
  e->pos_in_delay_left--;
  e->pos_in_delay_left&=255;

  for (i = 0; i < e->fir_length; i++) {
    ret += e->delay_left[cur] * e->fir[i];
    cur++;
    cur&=255;
  }

  return ret;
}

static double do_fir_right(eq *e, double sample)
{
  int i;
  int cur = e->pos_in_delay_right;
  double ret = 0;

  e->delay_right[e->pos_in_delay_right] = sample;
  e->pos_in_delay_right--;
  e->pos_in_delay_right&=255;

  for (i = 0; i < e->fir_length; i++) {
    ret += e->delay_right[cur] * e->fir[i];
    cur++;
    cur&=255;
  }

  return ret;
}

//#define FFT_LENGTH	(65536)
#define FFT_LENGTH	(65536*2)
//#define FFT_LENGTH	(2048*2)

#define ERR(x)		\
  do {			\
    err = x;		\
    goto code_err;	\
  } while (0)

static void feed_display_buffer(complex *in, int *out)
{
  /* in input we have a buffer with N frequencies.
   * In output we want M frequencies (with M < N).
   * Except we don't want to put N/M frequencies per cell
   * in the output buffer. We want a log scale. The first
   * cells take a few cells from the input, the last cells
   * take a lot. We want an 'octave' display, which is
   * more natural regarding music.
   * Thus the math that follows:
   * How to translate a range of freqs from input to output?
   * Here is the curve we want:
   * input freq
   * ^                .
   * |                .
   * |               .
   * |              .
   * |            .
   * |        .
   * |   .
   * ------------------------------> output freq
   * Why that?
   * Because we want the first output cells to represent a few input,
   * so the curve starts "flatty." Later on, for a little range in the
   * output, we want a lot of freqs from the input, thus a more vertical
   * curve at this point.
   * So we want some kind of exp curve. (By the way since we want
   * logarithmic stuff, it's no surprise to choose an exp curve, is it?)
   * So our function, that maps an output freq to an input one is
   * something like:
   * f(x) = 2^(kx) * r
   * Two parameters to set: k and r.
   * We are interested in audio freqs, so 20-20000Hz.
   * To make things simpler, let's start from 20 Hz and go up 10 octaves.
   * Meaning we calculate: 20 * 2^10 = 20480
   * Our 10 octaves give use: 20-20480Hz.
   * Let's arbitrarily set:
   * f(0) = 20
   * f(256) = 20480
   * If we calculate
   * f(x) = 2^(kx) * r
   * at point 0, we find:
   * 20 = 2^(k*0) * r
   * thus:
   * r = 20
   * at point 256:
   * 20480 = 2^(256k) * 20
   * 1024 = 2^(256k)
   * log2(1024) = 256k
   * k = 10/256
   * And there we are.
   * In reality we don't start at freq 20Hz but 0 and we end at 22050,
   * not 20480. What to do with the outter freqs? Well, let's integrate
   * them in the first and last point, and live with that. Not very
   * correct, but hey...
   * Then with the input freq we get the point in the array by simple
   * math:
   * 0 -> 0
   * y -> x
   * fft_length/2 -> 22050
   * y / (fft_length/2) = x/22050
   * y = x / 22050 * (fft_length/2)
   * and there you are (yes we use only the first half of the input buffer
   * because it's fft stuff, get some info on it to get the point)
   */
  double k = 10./256.;
  double r = 20;
  int i;

  for (i = 0; i < 256; i++) {
    int j;
    double v_re, v_im;
    double v;
    /* we don't want input buffer ranges to overlap. So we always use
     * floor. The min of the current range is equal to the max of the
     * previous range + 1
     */
    int min_pos_in_fft = floor(pow(2,k*i)*r * (FFT_LENGTH / 2.) / 22050.)+1;
    int max_pos_in_fft = floor(pow(2,k*(i+1))*r * (FFT_LENGTH / 2.) / 22050.);
//min_pos_in_fft = floor((double)i/256. * FFT_LENGTH/2.) + 1;
//max_pos_in_fft = floor((double)(i+1)/256. * FFT_LENGTH/2.);
    /* min might be > max so let's check */
    if (max_pos_in_fft < min_pos_in_fft) max_pos_in_fft = min_pos_in_fft;
    /* we take care of first and last range */
    if (i==0) min_pos_in_fft = 0;
    if (i==255) max_pos_in_fft = FFT_LENGTH/2-1;
    /* now what? Well, a simple average */
    v_re = 0;
    v_im = 0;
    for (j = min_pos_in_fft; j <= max_pos_in_fft; j++) {
      v_re += in[j].re;
      v_im += in[j].im;
    }
    v_re /= (max_pos_in_fft-min_pos_in_fft)+1;
    v_im /= (max_pos_in_fft-min_pos_in_fft)+1;
    /* normally it's a real fft, no need to take the imaginary part,
     * but just to be sure... (heck, I'm such a lamah!)
     */
    v = sqrt(v_re * v_re + v_im * v_im);
    /* might be 0, why not? so let's pretend it is -80 dB */
    if (v == 0) v = -80;
    else v = 20 * log10(v);
    /* and map to 0-100:
     * 0 -> 20
     * y -> x
     * 100 -> -80
     * y/100 = (x-20)/(-80-20)
     * y = -x + 20
     */
    out[i] = round(-v + 20);
  }
}

/*
 * list of commands:
 * fir_length <length>
 * range <f min> <f max> <gain (dB)>
 * add <f min> <f max> <gain (dB)>
 * eq10 <gain 1 (dB)> ... <gain 10 (dB)>
 * window <window type>
 *  <window type> is one of:
 *    square, blackman, bartlett, hamming, hanning
 *    no understood (thus not implemented): kaiser, lanczos, tukey
 * a command must end with ';'
 * comments start by '#' and extend to the end of the line
 */
void run_code_eq(eq *e)
{
  char *err = NULL;
  complex *wanted;
  complex *ifft;
  double *impulse;
  fft f;
  int i;

  hide_error_eq(e);

  /* we do a big fft to have enough resolution at low freqs
   * (we display logarithmically on the x axis, for an octave
   * to have the same width, meaning from 20 to 40 Hz you have
   * the same width than from 40 to 80 Hz, 80 to 160 Hz, 160 to 320, etc.)
   */
  wanted = malloc(sizeof(complex) * FFT_LENGTH);
  ifft = malloc(sizeof(complex) * FFT_LENGTH);

  if (!wanted || !ifft || open_fft(&f, FFT_LENGTH) == -1) goto run_err;

  /* some initial values */
  e->fir_length = 8;
  e->window_type = EQ_SQUARE;

  /* let's say we want all frequencies */
  for (i = 0; i < FFT_LENGTH/2; i++) {
    wanted[i].re = 1;
    wanted[i].im = 0;
  }

  /* run the code (if there is one) */
  if (e->code) {
    int pos = 0;
    int len = strlen(e->code);

    while (pos < len) {
      /* remove leading spaces */
      while (pos < len && isspace(e->code[pos])) pos++;
      if (pos == len) break;

      /* remove comments */
      if (e->code[pos] == '#') {
        while (pos < len && e->code[pos] != '\n') pos++;
        pos++;
        continue;
      }

      /* look for known commands */
      if (!strncmp(e->code+pos, "range", 5)) {
        /* range command waits 3 floats */
        double f1, f2, val;
        int min, max;
        if (sscanf(e->code+pos+5, "%lg %lg %lg", &f1, &f2, &val) != 3)
          ERR("Bad use of range.\nDo: range <f1 (Hz)> <f2 (Hz)> <val (dB)>");
        /* some checks */
        if (f1 < 0 || f1 >= 22050) ERR("Bad value for first frequency");
        if (f2 < 0 || f2 > 22050) ERR("Bad value for second frequency");
        if (f1 > f2) ERR("First frequency must be < to second frequency");
        if (val < -80 || val > 20)
          ERR("Bad use of range.\nBad value for val.");
        /* from freq to array index */
        /* 0   x  22050
         * 0   y  FFT_LENGTH/2
         * x/22050 = y / (FFT_LENGTH/2)
         * y = x * FFT_LENGTH / 2 / 22050
         */
        min = f1 * FFT_LENGTH / 2 / 22050.;
        max = f2 * FFT_LENGTH / 2 / 22050.;
        /* just in case... (paranoid humanoid) (fuck this rock band by the
         * way) (fuck I forgot its name) (nobody cares...)
         */
        if (min > FFT_LENGTH/2-1) min = FFT_LENGTH/2-1;
        if (max > FFT_LENGTH/2-1) max = FFT_LENGTH/2-1;
        /* val is in dB, get the linear shit out of it
         * y = 20 log10 (x)
         * x = 10^(y/20)
         */
        val = pow(10, val/20.);
        for (i = min; i <= max; i++) wanted[i].re = val;
      } else

      if (!strncmp(e->code+pos, "add", 3)) {
        /* inc command waits 3 floats */
        double f1, f2, val;
        int min, max;
        if (sscanf(e->code+pos+3, "%lg %lg %lg", &f1, &f2, &val) != 3)
          ERR("Bad use of add.\nDo: add <f1 (Hz)> <f2 (Hz)> <val (dB)>");
        /* some checks */
        if (f1 < 0 || f1 >= 22050) ERR("Bad value for first frequency");
        if (f2 < 0 || f2 > 22050) ERR("Bad value for second frequency");
        if (f1 > f2) ERR("First frequency must be < to second frequency");
        if (val < -80 || val > 20)
          ERR("Bad use of add.\nBad value for val.");
        /* from freq to array index */
        /* 0   x  22050
         * 0   y  FFT_LENGTH/2
         * x/22050 = y / (FFT_LENGTH/2)
         * y = x * FFT_LENGTH / 2 / 22050
         */
        min = f1 * FFT_LENGTH / 2 / 22050.;
        max = f2 * FFT_LENGTH / 2 / 22050.;
        /* just in case... (paranoid humanoid) (fuck this rock band by the
         * way) (fuck I forgot its name) (nobody cares...)
         */
        if (min > FFT_LENGTH/2-1) min = FFT_LENGTH/2-1;
        if (max > FFT_LENGTH/2-1) max = FFT_LENGTH/2-1;
        /* val is in dB, get the linear shit out of it
         * y = 20 log10 (x)
         * x = 10^(y/20)
         */
        val = pow(10, val/20.);
        for (i = min; i <= max; i++) wanted[i].re *= val;
      } else

      if (!strncmp(e->code+pos, "fir_length", 10)) {
        /* fir_length command waits 1 int */
        int f;
        if (sscanf(e->code+pos+10, "%d", &f) != 1)
          ERR("Bad use of fir_length.\nDo: fir_length <n>");
        if (f < 2 || f > 256)
          ERR("Bad use of fir_length.\nn must respect: 2 <= n <= 256");
        e->fir_length = f;
      } else

      if (!strncmp(e->code+pos, "eq10", 4)) {
        /* 10 band eq - freqs are:
         * 31.5, 63, 125, 250, 500, 1000, 2000, 4000, 8000 and 16000 Hz
         */
        /* note that we start from 0 Hz for the first band.
         * note also that the last band goes up to 22050 Hz.
         * What to do with first and last frequencies by the way?
         */
        /* so we wait for 10 numbers */
        double val[10];
        /* these are the centers as defined by ISO or something like that */
        double freqs[10] = { 31.5, 63, 125, 250, 500, 1000, 2000,
                             4000, 8000, 16000 };
        double min_freq, max_freq;
        /* mid_factor: do f * mid_factor to get the freq at distance
         * 1/2 octave higher than f
         */
        double mid_factor = pow(2, 1/2);
        int min, max;
        int i;
        if (sscanf(e->code+pos+4, "%lg %lg %lg %lg %lg %lg %lg %lg %lg %lg",
                   &val[0], &val[1], &val[2], &val[3], &val[4],
                   &val[5], &val[6], &val[7], &val[8], &val[9]) != 10)
          ERR("Bad use of eq10.\nDo: eq10 <val1 (dB)> .. <val10 (dB)>");
        for (i = 0; i < 10; i++)
          if (val[i] < -80 || val[i] > 20)
            ERR("Bad use of eq10.\nWe must have: -80 <= val <= 20");
        /* let's feed the buffer */
        for (i = 0; i < 10; i++) {
          int j;
          if (i == 0)
            min_freq = 0;
          else
            min_freq = freqs[i-1] * mid_factor;
          if (i == 9)
            max_freq = 22050;
          else
            max_freq = freqs[i] * mid_factor;
          /* from freq to array index */
          min = min_freq * FFT_LENGTH / 2 / 22050.;
          max = max_freq * FFT_LENGTH / 2 / 22050.;
          /* sanity checks (maybe useless) */
          if (min > FFT_LENGTH/2-1) min = FFT_LENGTH/2-1;
          if (max > FFT_LENGTH/2-1) max = FFT_LENGTH/2-1;
          /* feed my ass */
          val[i] = pow(10, val[i] / 20.);
          /* maybe the limit freqs between bands are not correct
           * (a freq belonging to one band may get the value of
           * the band near it because I don't know exactly how
           * min_freq and max_freq are truncated, but heck...
           * anyway later on the FIR filter will fuck it all up, so...
           */
          for (j = min; j <= max; j++) wanted[j].re = val[i];
        }
      } else
      if (!strncmp(e->code+pos, "window", 6)) {
        /* window */
        pos+=6;
        while (pos < len && isspace(e->code[pos])) pos++;
        if (pos == len)
          ERR("Bad use of window.");
        e->window_type = -1;
        if (!strncmp(e->code+pos, "square", 6))
          e->window_type = EQ_SQUARE;
        else if (!strncmp(e->code+pos, "bartlett", 8))
          e->window_type = EQ_BARTLETT;
        else if (!strncmp(e->code+pos, "blackman", 8))
          e->window_type = EQ_BLACKMAN;
        else if (!strncmp(e->code+pos, "hamming", 7))
          e->window_type = EQ_HAMMING;
        else if (!strncmp(e->code+pos, "hanning", 7))
          e->window_type = EQ_HANNING;
        if (e->window_type == -1)
          ERR("Bad use of window. Unknown window's type.");
      } else
        /* no known command? error */
        ERR("Unknown command.");

      /* get the trailing ';' */
      while (pos < len && e->code[pos] != ';') pos++;
      if (pos == len || e->code[pos] != ';') ERR("Unterminated command.");
      pos++;
    }
  }

  /* feed displayed wanted buffer */
  feed_display_buffer(wanted, e->gui.wanted_values);

  /* set the symmetry in the wanted data (to get real (ie. not complex)
   * ifft or something like that) (i don't understand that shit well...)
   */
  for (i = FFT_LENGTH / 2 + 1; i < FFT_LENGTH; i++) {
    wanted[i].re = wanted[FFT_LENGTH-i].re;
    wanted[i].im = -wanted[FFT_LENGTH-i].im;
  }
  /* I really don't understand this crap, but in the book, there is
   * this formula for a real sequence x(n):
   * X(N-k) = X*(k) = X(-k)
   * so let's use that
   * we will for X(N/2) = 0
   */
#if 0
  wanted[FFT_LENGTH/2].re = 0;
  wanted[FFT_LENGTH/2].im = 0;
  for (i = 1; i < FFT_LENGTH/2; i++) {
    wanted[FFT_LENGTH-i].re = wanted[i].re;
    wanted[FFT_LENGTH-i].im = wanted[i].im;
  }
#endif
#if 0
  for (i = FFT_LENGTH / 2; i < FFT_LENGTH; i++) {
    wanted[i].re = 0;
    wanted[i].im = 0;
  }
#endif

  /* calc the ifft */
  ifft_double(&f, wanted, ifft);

  /* pick up the wanted number of samples and put them in the
   * fir array
   */
  /* the following line is a hack but should work everywhere, I mean a complex
   * is made of two doubles
   */
  impulse = (double *)wanted;
#if 1
  {
    int odd = e->fir_length & 1;
    int len = e->fir_length / 2;
    /* we get the impulse from the ifft array
     * first data from the end followed by data from the start
     * something like that...
     * it's a "window" fir design method (except it surely sucks
     * like hell the way I do it...)
     */
    for (i = 0; i < len; i++)
      impulse[i] = ifft[i+FFT_LENGTH-len].re;
    for (i = 0; i < len + odd; i++)
      impulse[i+len] = ifft[i].re;
    for (i = e->fir_length; i < FFT_LENGTH; i++) impulse[i] = 0;
  }

  /* window the impulse */
  switch (e->window_type) {
    case EQ_SQUARE:
      /* nothing to do */
      break;
    case EQ_BARTLETT:
      for (i=0; i<e->fir_length; i++)
        impulse[i] *= 1 - (2 * fabs((double)i - (e->fir_length-1.)/2.)) /
                          ((double)e->fir_length-1);
      break;
    case EQ_BLACKMAN:
      for (i=0; i<e->fir_length; i++)
        impulse[i] *= 0.42 - 0.5 * cos(2 * M_PI * i / (e->fir_length-1)) +
                      0.08 * cos(4 * M_PI * i / (e->fir_length-1));
      break;
    case EQ_HAMMING:
      for (i=0; i<e->fir_length; i++)
        impulse[i] *= 0.54 - 0.46 * cos(2 * M_PI * i / (e->fir_length-1));
      break;
    case EQ_HANNING:
      for (i=0; i<e->fir_length; i++)
        impulse[i] *= 0.5 * (1 - cos(2 * M_PI * i / (e->fir_length-1)));
      break;
  }

  /* store the result in our structure */
  for (i = 0; i < e->fir_length; i++)
    e->fir[i] = impulse[i];

#endif
#if 0
  for (i = 0; i < e->fir_length; i++)
    impulse[i] = sqrt(ifft[i].re*ifft[i].re+ifft[i].im*ifft[i].im);
  for (i = e->fir_length; i < FFT_LENGTH; i++) impulse[i] = 0;
#endif

#if 0
for (i = 0; i < FFT_LENGTH; i++) impulse[i] = 0;
  impulse[0] = impulse[14] = -0.014112893;
  impulse[1] = impulse[13] = -0.001945309;
  impulse[2] = impulse[12] = 0.04000004;
  impulse[3] = impulse[11] = 0.01223454;
  impulse[4] = impulse[10] = -0.09138802;
  impulse[5] = impulse[9] = -0.01808986;
  impulse[6] = impulse[8] = 0.3133176;
  impulse[7] = 0.52;
#endif

  /* calc the fft of the selected impulse */
  fft_double(&f, impulse, ifft);

  /* feed displayed obtained buffer */
  feed_display_buffer(ifft, e->gui.obtained_values);

  if (e->c.gui_on)
    force_redraw_eq(e);

  close_fft(&f);
  free(wanted);
  free(ifft);

  return;

run_err:
  if (wanted) free(wanted);
  if (ifft) free(ifft);
  /* no need to close fft, there was an error, in every case
   * you don't have to close it (do the thinking, you will see...)
   * (go read fft.c to see when open_fft returns -1)
   */
  oom();
  display_error();
  return;

code_err:
  close_fft(&f);
  free(wanted);
  free(ifft);
  show_error_eq(e, err);
}

static int feed_buffer_float_mono(eq *this, float *buf, int nb_samples, int start)
{
  int i;

  if (this->c.bypass)
    return this->c.subchannel->c.feed_buffer_float_mono(this->c.subchannel,
        buf, nb_samples, start);

  if (!this->c.subchannel->c.feed_buffer_float_mono(this->c.subchannel,
        buf, nb_samples, start))
   return 0;

  for (i = 0; i < nb_samples; i++)
    buf[i] = do_fir_left(this, buf[i]);

  return 1;
}

static int feed_buffer_float_stereo(eq *this, float *buf, int nb_samples, int start)
{
  int i;

  if (this->c.bypass)
    return this->c.subchannel->c.feed_buffer_float_stereo(this->c.subchannel,
        buf, nb_samples, start);

  if (!this->c.subchannel->c.feed_buffer_float_stereo(this->c.subchannel,
        buf, nb_samples, start))
    return 0;

  for (i = 0; i < nb_samples * 2; i++) {
    buf[i] = do_fir_left(this, buf[i]);
    i++;
    buf[i] = do_fir_right(this, buf[i]);
  }

  return 1;
}

static int first_point(eq *this)
{
  return this->c.subchannel->c.first_point(this->c.subchannel);
}

static int last_point(eq *this)
{
  return this->c.subchannel->c.last_point(this->c.subchannel);
}

static void reset(eq *this)
{
  int i;

  this->c.subchannel->c.reset(this->c.subchannel);

  if (save_code_from_gui(this))
    run_code_eq(this);

  this->pos_in_delay_left = 0;
  this->pos_in_delay_right = 0;

  for (i = 0; i < 256; i++) {
    this->delay_left[i] = 0;
    this->delay_right[i] = 0;
  }
}

eq *new_eq(channel *c)
{
  int i;
  eq *ret = malloc(sizeof(eq));

  if (!ret)
    return 0;

  memset(ret, 0, sizeof(eq));

  /* by default fir length of 8 and filter does nothing */
  ret->fir_length = 8;
  ret->fir[0] = 1;
  for (i = 1; i < 8; i++) ret->fir[i] = 0;

  ret->c.type = CHANNEL_TYPE_EQ;

  ret->c.subchannel = c;

  ret->c.feed_buffer_float_mono = (int (*)(void *, float *, int, int))feed_buffer_float_mono;
  ret->c.feed_buffer_float_stereo = (int (*)(void *, float *, int, int))feed_buffer_float_stereo;
  ret->c.first_point = (int (*)(void *))first_point;
  ret->c.last_point = (int (*)(void *))last_point;
  ret->c.reset = (void (*)(void *))reset;

  ret->c.open_gui = (void (*)(void *, int))eq_open_gui;
  ret->c.close_gui = (void (*)(void *))eq_close_gui;

  /* at init, wanted_values and obtained_values are at 20 */
  for (i=0; i<256; i++) {
    ret->gui.wanted_values[i] = 20;
    ret->gui.obtained_values[i] = 20;
  }

  return ret;
}

void delete_eq(eq *e)
{
  if (e->code) free(e->code);
  free(e);
}
