/*
 * songs - a little tool to record and mix audio files.
 *
 * public domain
 *
 * http://sed.free.fr/songs
 */

#include "memmap.h"

#include <sys/stat.h>
#include <unistd.h>
#ifndef WIN
#include <sys/mman.h>
#else
#include <windows.h>
#include <stdio.h>
#endif
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#include "error.h"

typedef struct {
  void *buf;
  dev_t device;
  ino_t inode;
  int count;
#ifdef WIN
  char *name;
  HANDLE file;
  HANDLE map;
#endif
} map_list;

map_list *list = 0;
int size = 0;
int maxsize = 0;

/* like mmap, returns -1 if error */
void *memmap(struct stat *buf, int f, char *name)
{
  int i;
  void *b;

  /* search the list */
#ifndef WIN
  for (i=0; i<size; i++)
    if (buf->st_dev == list[i].device && buf->st_ino == list[i].inode) {
      list[i].count++;
      return list[i].buf;
    }
#else
  for (i=0; i<size; i++)
    if (list[i].name && !strcmp(list[i].name, name)) {
      list[i].count++;
      return list[i].buf;
    }
#endif

  /* not there */
  if (size == maxsize) {
    map_list *n = realloc(list, (maxsize + 32) * sizeof(map_list));
    if (!n) {
      oom();
      display_error();
      return (void *)-1;
    }
    list = n;
    maxsize += 32;
  }

#ifndef WIN
  b = mmap(0, buf->st_size, PROT_READ, MAP_SHARED, f, 0);
#else
  {
    HANDLE file;
    HANDLE map;
    DWORD file_size;
    LPVOID map_address;

    file = CreateFile(name, GENERIC_READ, FILE_SHARE_READ, NULL,
                      OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
    if (file == INVALID_HANDLE_VALUE) {
      printf("error 1\n");
      return (void *)-1;
    }

    file_size = GetFileSize(file, NULL);

    map = CreateFileMapping(file, NULL, PAGE_READONLY, 0, file_size, NULL);
    if (map == NULL) {
      printf("error 1\n");
      CloseHandle(file);
      return (void *)-1;
    }

    map_address = MapViewOfFile(map, FILE_MAP_READ, 0, 0, file_size);
    if (map_address == NULL) {
      printf("error 2\n");
      CloseHandle(file);
      CloseHandle(map);
      return (void *)-1;
    }

    b = (void *)map_address;

    list[size].file = file;
    list[size].map = map;
  }
#endif

  if (b == (void *)-1) {
    int serr = errno;
    reset_error();
    add_error_char(name);
    add_error_char(":");
    add_error_char(strerror(serr));
    add_error_char("\n");
    display_error();
    return (void *)-1;
  }

  list[size].buf = b;
  list[size].device = buf->st_dev;
  list[size].inode = buf->st_ino;
  list[size].count = 1;

#ifdef WIN
  /* we accept failure of strdup. If later we map the file again,
   * it will be mapped twice. Not our problem.
   */
  list[size].name = strdup(name);
#endif

  size++;

  return b;
}

void memunmap(void *b, int s)
{
  int i;

  /* it has to be in there */
  for (i=0; i<size; i++)
    if (list[i].buf == b)
      break;

  if (i == size) {
    reset_error();
    add_error_char(__FILE__);
    add_error_char(":");
    add_error_int(__LINE__);
    add_error_char(":");
    add_error_char((char *)__FUNCTION__);
    add_error_char(": you should not read this, contact the authors.\n");
    display_error();
    return;
  }

  list[i].count--;

  if (list[i].count)
    return;

#ifndef WIN
  munmap(b, s);
#else
  if (list[i].name) free(list[i].name);
  UnmapViewOfFile(list[i].buf);
  CloseHandle(list[i].map);
  CloseHandle(list[i].file);
#endif

  /* we have to remove it from the list */
  for (; i<size-1; i++)
    list[i] = list[i+1];

  size--;

  /* make some room */
  if (maxsize-size > 32) {
    map_list *new = realloc(list, (maxsize-32) * sizeof(map_list));
    /* if it does not work, no problem */
    if (!new)
      return;
    maxsize -= 32;
    list = new;
  }
}

void cleanup_map(void)
{
  if (size) {
    reset_error();
    add_error_char(__FILE__);
    add_error_char(":");
    add_error_int(__LINE__);
    add_error_char(":");
    add_error_char((char *)__FUNCTION__);
    add_error_char(": you should not read this, contact the authors.\n");
    display_error();
  }

  if (!maxsize)
    return;

  /* we should free names under windows. We don't care, it's at end of
   * program, not a big issue
   */

  free(list);
  list = 0;
  maxsize = 0;
}
