/*
 * songs - a little tool to record and mix audio files.
 *
 * public domain
 *
 * http://sed.free.fr/songs
 */

#ifndef WIN

#include "soundcard.h"

#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "error.h"
#include "setup_gui.h"

int fragment_size[11]={         /* 7, ..., 17 makes 11 values */
  /* fragment size is in [7..17] says mikmod 3.0.3, so... */
  128, 256, 512, 1024, 2048, 4096, 8192, 16384, 32768, 65536, 131072
};

/* the audio device file handle */
static int soundd;

static int inited=0;

void sound_stats(void)
{
#if 0
  struct audio_buf_info in, out;
  int cap;
  if (ioctl(soundd, SNDCTL_DSP_GETOSPACE, &out) == -1 ||
      ioctl(soundd, SNDCTL_DSP_GETISPACE, &in)  == -1) {
    perror("ioctl");
    return;
  }
  fprintf(stdout, "%d %d in: %d %d %d %d out : %d %d %d %d\n",
    in.fragsize, out.fragsize,
    in.fragments, in.fragstotal, in.fragsize, in.bytes,
    out.fragments, out.fragstotal, out.fragsize, out.bytes);

  if (ioctl(soundd, SNDCTL_DSP_GETCAPS, &cap) == -1) {
    perror("ioctl");
    return;
  }
  fprintf(stdout, "%d\n", !!(cap & DSP_CAP_BATCH));
#endif

  struct count_info in, out;
  if (ioctl(soundd, SNDCTL_DSP_GETOPTR, &out) == -1 ||
      ioctl(soundd, SNDCTL_DSP_GETIPTR, &in)  == -1) {
    perror("ioctl");
    return;
  }

  fprintf(stdout, "in: bytes = %d blocks = %d ptr = %d\nout: bytes = %d blocks = %d ptr = %d\n",
     in.bytes, in.blocks, in.ptr,
     out.bytes, out.blocks, out.ptr);

  fprintf(stdout, "diff = %d\n", in.bytes - out.bytes * 2);
}

int get_sound_fd(void)
{
  if (!inited)
    return -1;

  return soundd;
}

/* return 0 if full (or error) 1 if not */
int not_full_sound(void)
{
  audio_buf_info info;

  if (ioctl(soundd, SNDCTL_DSP_GETOSPACE, &info)==-1)
    return 0;

  return info.fragments!=0;
}

/* this function sets the card output resolution, ie 8/16 bits, mono(0)/stereo(1), sampling rate (8000, 44100, ...) */
/* return 0 on error, 1 on success */
int set_sound_type(int bit_resolution, int mono_stereo, int sampling_rate, int fragsize)
{
  int numfrags=9;
  int s0, s1, s2;
  int fragmentsize;

  s0 = fragmentsize = ((numfrags<<16) | fragsize);

  if(ioctl(soundd, SNDCTL_DSP_SETFRAGMENT, &fragmentsize)<0) {
    int serr = errno;
    reset_error();
    add_error_char("Audio\n");
    add_error_char("ioctl on ");
    add_error_char(audio_file);
    add_error_char(": ");
    add_error_char(strerror(serr));
    display_error();
    return 0;
  }

  if (s0 != fragmentsize)
    goto not_supported;

  s0 = bit_resolution;
  s1 = mono_stereo;
  s2 = sampling_rate;

  if (ioctl(soundd, SNDCTL_DSP_SAMPLESIZE, &bit_resolution)==-1 ||
      ioctl(soundd, SNDCTL_DSP_STEREO, &mono_stereo)==-1 ||
      ioctl(soundd, SNDCTL_DSP_SPEED, &sampling_rate)==-1) {
    int serr = errno;
    reset_error();
    add_error_char("Audio\n");
    add_error_char("ioctl on ");
    add_error_char(audio_file);
    add_error_char(": ");
    add_error_char(strerror(serr));
    display_error();
    return 0;
  }

  if (s0 != bit_resolution || s1 != mono_stereo || s2 != sampling_rate)
    goto not_supported;

  ioctl(soundd, SNDCTL_DSP_RESET, 0);

  return 1;

not_supported:
  reset_error();
  add_error_char("Audio\n");
  add_error_char("The requested audio mode is not supported by your soundcard.\n");
  display_error();
  return 0;
}

/* return 0 if ok, -1 otherwise */
int unblock_sound(int fd)
{
  int fl;

  if ((fl=fcntl(fd, F_GETFL))==-1) {
    perror("card.unblock_sound.fcntl1");
    return -1;
  }
  fl|=O_NONBLOCK;
  if ((fl=fcntl(fd, F_SETFL, fl))==-1) {
    perror("card.unblock_soud.fcntl2");
    return -1;
  }
  return 0;
}

/* return 0 on error 1 otherwise*/
int init_sound(int sample_rate, int precision, int mono_stereo, int fragsize, int type)
{
  int mode = 0;

  if (inited==1) {
    reset_error();
    add_error_char("Audio\n");
    add_error_char("Audio already in use inside Songs\n");
    display_error();
    return 0;
  }

  if (type == PLAY_SOUND)
    mode = O_WRONLY;
  else if (type == RECORD_SOUND)
    mode = O_RDONLY;
  else if (type == PLAY_AND_RECORD_SOUND)
    mode = O_RDWR;
  else {
    reset_error();
    add_error_char("Audio\n");
    add_error_char("Bad mode for the audio stuff: ");
    add_error_int(type);
    add_error_char("\n");
    display_error();
    return 0;
  }

  soundd=open(audio_file, mode, 0);
  if (soundd == -1) {
    int serr = errno;
    reset_error();
    add_error_char("Audio\n");
    add_error_char(audio_file);
    add_error_char(": ");
    add_error_char(strerror(serr));
    if (type == PLAY_AND_RECORD_SOUND) {
      add_error_char("(Your soundcard probably does not support full duplex\n");
      add_error_char("or there is no driver for it in your operating system)\n");
    }
    display_error();
    return 0;
  }

  if (!set_sound_type(precision, mono_stereo, sample_rate, fragsize) /*|| unblock_sound(soundd)==-1*/) {
    close(soundd);
    return 0;
  }

  inited=1;

  return 1;
}

void close_sound(void)
{
  if (!inited)
    return;

  ioctl(soundd, SNDCTL_DSP_RESET, 0);
  close(soundd);
  inited=0;
}

/* return 0 if problem, 1 if ok */
int play_sound(void *buf, int buf_len)
{
  int serr;

  if (!inited)
    return 0;
  if (write(soundd, buf, buf_len)!=buf_len) {
    serr = errno;
    reset_error();
    add_error_char("Audio\n");
    add_error_char("write:");
    add_error_char(strerror(serr));
    display_error();
    close_sound();
    return 0;
  }
  return 1;
}

/* return 0 if problem, 1 if ok */
int record_sound(void *buf, int buf_len)
{
  int serr;

  if (!inited)
    return 0;
  if (read(soundd, buf, buf_len)!=buf_len) {
    serr = errno;
    reset_error();
    add_error_char("Audio\n");
    add_error_char("read:");
    add_error_char(strerror(serr));
    display_error();
    close_sound();
    return 0;
  }
  return 1;
}

#else /* WIN */

#include "soundcard.h"

#include <windows.h>
#include <gtk/gtk.h>

/* this code is VERY convoluted. Don't even try to understand.
 * I hate windows, what do you expect, huh?
 * But I need Songs to work under windows because that's all
 * I have under my hand for the moment and I have a song to record.
 */

/* if something fails, we are in deep shit, errors are not well handled. */

static int sound_type = -1;

#define NB_BUFFERS 5

/* length is hardcoded to 4096 */
static WAVEHDR play_wavehdr[NB_BUFFERS];
static char play_buffer[NB_BUFFERS][4096];
static volatile int play_busy[NB_BUFFERS];
static HWAVEOUT out;

static long my_thread_id;

/* return 0 if err, 1 if ok */
int play_sound(void *buf, int buf_len)
{
  int i;
  int free_buffer = -1;

  for (i = 0; i < NB_BUFFERS; i++)
    if (play_busy[i] == 0) {
      free_buffer = i;
      break;
  }
  if (free_buffer == -1) return 0;

  memcpy(play_buffer[free_buffer], buf, 4096);
  play_busy[free_buffer] = 1;

  play_wavehdr[free_buffer].lpData = play_buffer[free_buffer];
  play_wavehdr[free_buffer].dwBufferLength = 4096;
  play_wavehdr[free_buffer].dwFlags = 0;
  play_wavehdr[free_buffer].dwLoops = 0;
  if (waveOutPrepareHeader(out, &play_wavehdr[free_buffer], sizeof(WAVEHDR)) !=
      MMSYSERR_NOERROR)
    printf("prepare out error!\n");
  if (waveOutWrite(out, &play_wavehdr[free_buffer], sizeof(WAVEHDR)) !=
      MMSYSERR_NOERROR)
    printf("write out error!\n");

  return 1;
}

static volatile int input_head; /* points to the next ready buffer */
static volatile int input_tail; /* points to the last ready buffer + 1 */
static WAVEHDR record_wavehdr[NB_BUFFERS];
static char record_buffer[NB_BUFFERS][4096];
static HWAVEIN in;

int record_sound(void *buf, int buf_len)
{
  memcpy(buf, record_buffer[input_head], 4096);

  waveInUnprepareHeader(in, &record_wavehdr[input_head], sizeof(WAVEHDR));
  record_wavehdr[input_head].lpData = record_buffer[input_head];
  record_wavehdr[input_head].dwBufferLength = 4096;
  record_wavehdr[input_head].dwFlags = 0;
  if (waveInPrepareHeader(in, &record_wavehdr[input_head],
                          sizeof(WAVEHDR)) != MMSYSERR_NOERROR) {
    printf("error prepare header\n");
    return 0;
  }
  if (waveInAddBuffer(in, &record_wavehdr[input_head],
                      sizeof(WAVEHDR)) != MMSYSERR_NOERROR) {
    printf("error add buffer\n");
    return 0;
  }

  input_head++;
  if (input_head == NB_BUFFERS) input_head = 0;

  return 1;
}

static void CALLBACK play_callback(HWAVEOUT out, UINT msg,
                                   DWORD *inst, DWORD *p1, DWORD *p2)
{
  switch (msg) {
    case WOM_OPEN:
    case WIM_OPEN:
      PostThreadMessage(my_thread_id, WM_USER, 0, 0);
      break;
    case WOM_DONE: {
      int buffer = (WAVEHDR *)p1 - play_wavehdr;
      if (buffer < 0 || buffer > NB_BUFFERS - 1) {
        printf("hum... problem... you don't read that, do you?\n");
        return;
      }
      play_busy[buffer] = 0;
      if (!PostThreadMessage(my_thread_id, WM_USER, 0, 0))
        printf("[post error] ");
      break;
    }
    case WIM_DATA:
      input_tail++;
      if (input_tail == NB_BUFFERS) input_tail = 0;
      if (!PostThreadMessage(my_thread_id, WM_USER, 0, 0))
        printf("[post error] ");
      break;
  }
}

gint (*system_poll)(GPollFD *f, guint n, gint t);
static gint my_poll(GPollFD *f, guint n, gint t);

void empty_queue(void)
{
  extern void (*play_ready)(void);
  extern void (*record_ready)(void);
  MSG m;
  int in_last = input_tail;

  while (PeekMessage(&m, (void*)-1, WM_USER, WM_USER, PM_REMOVE));

  while (input_head != in_last && record_ready)
    record_ready();

  if (play_ready) {
    int i;
    for (i = 0; i < NB_BUFFERS; i++)
      if (play_busy[i] == 0)
        play_ready();
  }
}

/* we hack the poll system of glib. It more or less works with
 * glib 2.16.5. Maybe it will miserably fail with other glib's
 * versions.
 * Here is how it works for the "play" mode:
 * the empty_queue function feeds as much buffers as available.
 * When a buffer is finished windows calls a callback.
 * This callback sends a MSG to the thread.
 * GTK has a g_poll function that waits for such messages.
 * It will return when one comes.
 * Now the problem:
 * the message sent is of type "ThreadMessage". I return that
 * to GTK which is not supposed to do anything with that and
 * doesn't expect it. Apparently it doesn't hurt. But maybe
 * it does (a crash might come at some point) after all, so
 * well...
 * I could try to avoid a return to GTK when only these MSG
 * have been processed by g_poll, but it's too much work. It
 * seems to work as it is, why change?
 */
#if 1
gint my_poll(GPollFD *f, guint n, gint t)
{
  gint ret;

  empty_queue();

  ret  = system_poll(f, n, t);

  empty_queue();

  return ret;
}
#endif

#if 0
static gboolean toto(GIOChannel *c, GIOCondition co, gpointer d)
{
  empty_queue();

  return TRUE;
}
#endif

int init_sound(int sample_rate, int precision, int mono_stereo,
               int fragsize, int type)
{
  MSG m;
  int i;
  WAVEFORMATEX format;

  PeekMessage(&m, NULL, WM_USER, WM_USER, PM_NOREMOVE);
  my_thread_id = GetCurrentThreadId();

  /* I tried to plug in a more "natural" way, but it was worse
   * especially when I move the window. And by the way, see the hack
   * I must do to force gtk to report messages associated with no
   * window? (this is glib 2.16.6, maybe we can attach thread messages
   * with more recent gtk).
   */
#if 0
  GIOChannel *c;
  c = g_io_channel_win32_new_messages(0);
  /* we need to say it's a readable channel */
  c->is_readable = 1;
  g_io_add_watch(c, G_IO_IN, toto, NULL);
#endif

  system_poll = g_main_context_get_poll_func(g_main_context_default());
  g_main_context_set_poll_func(g_main_context_default(), my_poll);

  sound_type = type;

  if (sound_type == PLAY_SOUND || sound_type == PLAY_AND_RECORD_SOUND) {
    format.wFormatTag = WAVE_FORMAT_PCM;
    format.nChannels = mono_stereo + 1;
    format.nSamplesPerSec = sample_rate;
    format.nAvgBytesPerSec = sample_rate * precision / 8 * (mono_stereo + 1);
    format.nBlockAlign = precision / 8 * (mono_stereo + 1);
    format.wBitsPerSample = precision;
    format.cbSize = 0;
    waveOutOpen(&out, WAVE_MAPPER, &format, (unsigned long)play_callback, 0,
                CALLBACK_FUNCTION);
    GetMessage(&m, (void*)-1, 0, 0);

    for (i = 0; i < NB_BUFFERS; i++)
      play_busy[i] = 0;
  }

  if (sound_type == RECORD_SOUND || sound_type == PLAY_AND_RECORD_SOUND ||
      sound_type == RECORD_ADJUST) {
    format.wFormatTag = WAVE_FORMAT_PCM;
    format.nChannels = mono_stereo + 1;
    format.nSamplesPerSec = sample_rate;
    format.nAvgBytesPerSec = sample_rate * precision / 8 * (mono_stereo + 1);
    format.nBlockAlign = precision / 8 * (mono_stereo + 1);
    format.wBitsPerSample = precision;
    format.cbSize = 0;
    if (waveInOpen(&in, WAVE_MAPPER, &format, (unsigned long)play_callback, 0,
                   CALLBACK_FUNCTION) != MMSYSERR_NOERROR) {
      printf("error open audio in\n");
      return 0;
    }
    GetMessage(&m, (void*)-1, 0, 0);

    input_head = input_tail = 0;

    for (i = 0; i < NB_BUFFERS; i++) {
      record_wavehdr[i].lpData = record_buffer[i];
      record_wavehdr[i].dwBufferLength = 4096;
      record_wavehdr[i].dwFlags = 0;
      if (waveInPrepareHeader(in, &record_wavehdr[i],
                              sizeof(WAVEHDR)) != MMSYSERR_NOERROR) {
        printf("error prepare header\n");
        return 0;
      }
      if (waveInAddBuffer(in, &record_wavehdr[i],
                          sizeof(WAVEHDR)) != MMSYSERR_NOERROR) {
        printf("error add buffer\n");
        return 0;
      }
    }

    if (waveInStart(in) != MMSYSERR_NOERROR) {
      printf("wave start input err\n");
      return 0;
    }
  }

  return 1;
}

void close_sound(void)
{
  int i;

  g_main_context_set_poll_func(g_main_context_default(), system_poll);

  if (sound_type == PLAY_SOUND || sound_type == PLAY_AND_RECORD_SOUND) {
    for (i = 0; i < NB_BUFFERS; i++)
      waveOutUnprepareHeader(out, &play_wavehdr[i], sizeof(WAVEHDR));
    waveOutClose(out);
  }

  if (sound_type == RECORD_SOUND || sound_type == PLAY_AND_RECORD_SOUND ||
      sound_type == RECORD_ADJUST) {
    for (i = 0; i < NB_BUFFERS; i++)
      waveInUnprepareHeader(in, &record_wavehdr[i], sizeof(WAVEHDR));
    waveInClose(in);
  }

  sound_type = -1;
}

#endif /* WIN */
