/*
 * sverb: an order 15 CFDN reverb
 * http://sed.free.fr/sverb
 * public domain
 * (there may be some patent for this stuff, I don't know and I don't care)
 * Cdric Roux <sed@free.fr>
 *
 * Algorithm comes from:
 * "Maximally Diffusive Yet Efficient Feedback Delay Networks for
 *  Artificial Reverberation", Davide Rocchesso, IEEE SIGNAL
 * PROCESSING LETTERS, VOL. 4, NO. 9, SEPTEMBER 1997
 * http://www.scienze.univr.it/documenti//ArticoloRivista/allegato/
 *   allegato128605.pdf
 * And:
 * "Circulant and elliptic feedback delay networks for artificial
 *  reverberation", D. Rocchesso and J. O. Smith, IEEE Trans. Speech
 * Audio Processing, vol. 5, pp. 51 63, Jan. 1997.
 * http://ccrma.stanford.edu/~jos/cfdn/cfdn.pdf
 * And:
 * http://ccrma.stanford.edu/~jos/Reverb/First_Order_Delay_Filter_Design.html
 * http://ccrma.stanford.edu/~jos/Reverb/Tonal_Correction_Filter.html
 * for the lowpass filters
 */

#ifndef _SVERB_H_
#define _SVERB_H_

#define SVERB_FLOAT	0
#define SVERB_INT	1
#ifdef ASM_IA32
#define SVERB_ASM	2
#endif

/* sverb types */
#define SVERB_TINY	0
#define SVERB_SMALL_1	1
#define SVERB_SMALL_2	2
#define SVERB_MEDIUM_1	3
#define SVERB_MEDIUM_2	4
#define SVERB_AVERAGE_1	5
#define SVERB_AVERAGE_2	6
#define SVERB_AVERAGE_3	7
#define SVERB_MIDDLE	8
#define SVERB_LARGE_1	9
#define SVERB_LARGE_2	10
#define SVERB_BIG_1	11
#define SVERB_BIG_2	12
#define SVERB_HUGE	13

/*
 * Tonal correction and loss delay stuff may be changed if you
 * have something better. In this case, it would be good
 * practice to share your stuff, but you do what you want.
 */
typedef struct {
  long c1;
  long c2;
  long oy;
} loss_delay_line_filter_int;

typedef struct {
  float c1;
  float c2;
  float oy;
} loss_delay_line_filter;

typedef struct {
  long c1;
  long c2;
  long ox;
} tonal_correction_filter_int;

typedef struct {
  float c1;
  float c2;
  float ox;
} tonal_correction_filter;

/* If you don't need integer or float, remove the part,
 * and adapt the code to remove functions dealing with
 * that.
 */
typedef struct sverb {
  /* this has to come first and in this order, it is used by the asm code */
  int DELAY[15];
  int cur[15];
  long *delays_int[15];
  loss_delay_line_filter_int ld_int[15];
  tonal_correction_filter_int tc_int;
  int integer_precision;
  /* end of asm code usage */
  float *delays[15];
  loss_delay_line_filter ld[15];
  tonal_correction_filter tc;
  float t60_0;
  float t60_pi;
  int type;
  int bypass_t60;
  float (*sverb_process)(struct sverb *, float);
} sverb;

/*
 * typical usage:
 * 
 * DELAY: array containing the delay lines
 * t60_0, t60_pi: the t60 parameters
 * x: array containing the input
 * This example will use the default float based code.
 * If you want integer or asm routine (if you have an IA32)
 * you need to call sverb_set_type with the good type.
 * In this case, you also can choose the integer precision
 * which is the number of bits used for the floating part
 * in the 32 bits integers. The default is 10.
 * If you want to bypass the t60 stuff, you must call
 * sverb_use_t60 with 0 as parameter.
 *
 * sverb s;
 * int i;
 * sverb_init(&s);
 * for (i=0; i<15; i++)
 *   if (sverb_set_delay(&s, i, DELAY[i]) == -1)
 *     return -1;
 * sverb_set_t60(t60_0, t60_pi);
 * sverb_reset(&s);
 * for (i=0; i<NB_SAMPLES; i++)
 *   s.sverb_process(&s, x[i]);
 *
 * Don't forget to process a bit of silence at the end,
 * so that the reverb has the time to finish. Otherwise,
 * you might experiment a brutal cut in the sound. The
 * listener will notice it and not appreciate it.
 */

/*
 * You must call it once and only once on the sverb structure.
 * It must come first, before you do anything with the sverb structure.
 */
void sverb_init(sverb *);

/*
 * reset the stuff, put everything to 0
 */
void sverb_reset(sverb *);

/*
 * free the mem malloced
 */
void sverb_free(sverb *);

/*
 * sverb_set_t60(s, t60_0, t60_pi)
 */
void sverb_set_t60(sverb *, float, float);

/*
 * sverb_set_type(s, type): type= SVERB_FLOAT, SVERB_INT or SVERB_ASM
 */
void sverb_set_type(sverb *, int);

/*
 * Sets the precision for the integer and asm version,
 * in number of bits for the floating part of the number.
 * set_integer_precision(s, precision)
 */
void sverb_set_integer_precision(sverb *, int);

/*
 * sverb_use_t60(s, yes_no): yes_no=0 for no t60 =1 for t60
 */
void sverb_use_t60(sverb *, int);

/*
 * sverb_set_delay(s, delay_line_number, delay_length):
 * return 0 if ok, -1 if error (bad delay line number, out of mem)
 */
int sverb_set_delay(sverb *, int, int);

#endif /* _SVERB_H_ */
